@@ -31,6 +31,8 @@
 # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 
+from __future__ import annotations
+
 import copy
 import csv
 import io
@@ -41,8 +43,9 @@ import re
 import textwrap
 from html import escape
 from html.parser import HTMLParser
+from typing import Any
 
-import wcwidth
+import wcwidth  # type: ignore
 
 # hrule styles
 FRAME = 0
@@ -59,6 +62,7 @@ ORGMODE = 14
 DOUBLE_BORDER = 15
 SINGLE_BORDER = 16
 RANDOM = 20
+BASE_ALIGN_VALUE = "base_align_value"
 
 _re = re.compile(r"\033\[[0-9;]*m|\033\(B")
 
@@ -71,8 +75,7 @@ def _get_size(text):
 
 
 class PrettyTable:
-    def __init__(self, field_names=None, **kwargs):
-
+    def __init__(self, field_names=None, **kwargs) -> None:
         """Return a new PrettyTable instance
 
         Arguments:
@@ -87,6 +90,8 @@ class PrettyTable:
         header_style - stylisation to apply to field names in header
             ("cap", "title", "upper", "lower" or None)
         border - print a border around the table (True or False)
+        preserve_internal_border - print a border inside the table even if
+            border is disabled (True or False)
         hrules - controls printing of horizontal rules after rows.
             Allowed values: FRAME, HEADER, ALL, NONE
         vrules - controls printing of vertical rules between columns.
@@ -129,8 +134,8 @@ class PrettyTable:
         self.encoding = kwargs.get("encoding", "UTF-8")
 
         # Data
-        self._field_names = []
-        self._rows = []
+        self._field_names: list[str] = []
+        self._rows: list[list] = []
         self.align = {}
         self.valign = {}
         self.max_width = {}
@@ -142,7 +147,7 @@ class PrettyTable:
         if field_names:
             self.field_names = field_names
         else:
-            self._widths = []
+            self._widths: list[int] = []
 
         # Options
         self._options = [
@@ -152,6 +157,7 @@ class PrettyTable:
             "fields",
             "header",
             "border",
+            "preserve_internal_border",
             "sortby",
             "reversesort",
             "sort_key",
@@ -188,6 +194,7 @@ class PrettyTable:
             "valign",
             "max_width",
             "min_width",
+            "none_format",
         ]
         for option in self._options:
             if option in kwargs:
@@ -199,6 +206,7 @@ class PrettyTable:
         self._start = kwargs["start"] or 0
         self._end = kwargs["end"] or None
         self._fields = kwargs["fields"] or None
+        self._none_format: dict[None, None] = {}
 
         if kwargs["header"] in (True, False):
             self._header = kwargs["header"]
@@ -209,6 +217,10 @@ class PrettyTable:
             self._border = kwargs["border"]
         else:
             self._border = True
+        if kwargs["preserve_internal_border"] in (True, False):
+            self._preserve_internal_border = kwargs["preserve_internal_border"]
+        else:
+            self._preserve_internal_border = False
         self._hrules = kwargs["hrules"] or FRAME
         self._vrules = kwargs["vrules"] or ALL
 
@@ -227,6 +239,7 @@ class PrettyTable:
         self.int_format = kwargs["int_format"] or {}
         self.float_format = kwargs["float_format"] or {}
         self.custom_format = kwargs["custom_format"] or {}
+        self.none_format = kwargs["none_format"] or {}
 
         self._min_table_width = kwargs["min_table_width"] or None
         self._max_table_width = kwargs["max_table_width"] or None
@@ -310,7 +323,7 @@ class PrettyTable:
         elif isinstance(index, int):
             new.add_row(self._rows[index])
         else:
-            raise Exception(f"Index {index} is invalid, must be an integer or slice")
+            raise IndexError(f"Index {index} is invalid, must be an integer or slice")
         return new
 
     def __str__(self):
@@ -342,6 +355,8 @@ class PrettyTable:
     def _validate_option(self, option, val):
         if option == "field_names":
             self._validate_field_names(val)
+        elif option == "none_format":
+            self._validate_none_format(val)
         elif option in (
             "start",
             "end",
@@ -368,6 +383,7 @@ class PrettyTable:
         elif option in (
             "header",
             "border",
+            "preserve_internal_border",
             "reversesort",
             "xhtml",
             "print_empty",
@@ -407,7 +423,7 @@ class PrettyTable:
             try:
                 assert len(val) == len(self._field_names)
             except AssertionError:
-                raise Exception(
+                raise ValueError(
                     "Field name list has incorrect number of values, "
                     f"(actual) {len(val)}!={len(self._field_names)} (expected)"
                 )
@@ -415,7 +431,7 @@ class PrettyTable:
             try:
                 assert len(val) == len(self._rows[0])
             except AssertionError:
-                raise Exception(
+                raise ValueError(
                     "Field name list has incorrect number of values, "
                     f"(actual) {len(val)}!={len(self._rows[0])} (expected)"
                 )
@@ -423,13 +439,22 @@ class PrettyTable:
         try:
             assert len(val) == len(set(val))
         except AssertionError:
-            raise Exception("Field names must be unique")
+            raise ValueError("Field names must be unique")
+
+    def _validate_none_format(self, val):
+        try:
+            if val is not None:
+                assert isinstance(val, str)
+        except AssertionError:
+            raise TypeError(
+                "Replacement for None value must be a string if being supplied."
+            )
 
     def _validate_header_style(self, val):
         try:
             assert val in ("cap", "title", "upper", "lower", None)
         except AssertionError:
-            raise Exception(
+            raise ValueError(
                 "Invalid header style, use cap, title, upper, lower or None"
             )
 
@@ -437,25 +462,25 @@ class PrettyTable:
         try:
             assert val in ["l", "c", "r"]
         except AssertionError:
-            raise Exception(f"Alignment {val} is invalid, use l, c or r")
+            raise ValueError(f"Alignment {val} is invalid, use l, c or r")
 
     def _validate_valign(self, val):
         try:
             assert val in ["t", "m", "b", None]
         except AssertionError:
-            raise Exception(f"Alignment {val} is invalid, use t, m, b or None")
+            raise ValueError(f"Alignment {val} is invalid, use t, m, b or None")
 
     def _validate_nonnegative_int(self, name, val):
         try:
             assert int(val) >= 0
         except AssertionError:
-            raise Exception(f"Invalid value for {name}: {val}")
+            raise ValueError(f"Invalid value for {name}: {val}")
 
     def _validate_true_or_false(self, name, val):
         try:
             assert val in (True, False)
         except AssertionError:
-            raise Exception(f"Invalid value for {name}. Must be True or False.")
+            raise ValueError(f"Invalid value for {name}. Must be True or False.")
 
     def _validate_int_format(self, name, val):
         if val == "":
@@ -464,7 +489,7 @@ class PrettyTable:
             assert isinstance(val, str)
             assert val.isdigit()
         except AssertionError:
-            raise Exception(
+            raise ValueError(
                 f"Invalid value for {name}. Must be an integer format string."
             )
 
@@ -483,19 +508,21 @@ class PrettyTable:
                 or (bits[1][-1] == "f" and bits[1].rstrip("f").isdigit())
             )
         except AssertionError:
-            raise Exception(f"Invalid value for {name}. Must be a float format string.")
+            raise ValueError(
+                f"Invalid value for {name}. Must be a float format string."
+            )
 
     def _validate_function(self, name, val):
         try:
             assert hasattr(val, "__call__")
         except AssertionError:
-            raise Exception(f"Invalid value for {name}. Must be a function.")
+            raise ValueError(f"Invalid value for {name}. Must be a function.")
 
     def _validate_hrules(self, name, val):
         try:
             assert val in (ALL, FRAME, HEADER, NONE)
         except AssertionError:
-            raise Exception(
+            raise ValueError(
                 f"Invalid value for {name}. Must be ALL, FRAME, HEADER or NONE."
             )
 
@@ -503,42 +530,42 @@ class PrettyTable:
         try:
             assert val in (ALL, FRAME, NONE)
         except AssertionError:
-            raise Exception(f"Invalid value for {name}. Must be ALL, FRAME, or NONE.")
+            raise ValueError(f"Invalid value for {name}. Must be ALL, FRAME, or NONE.")
 
     def _validate_field_name(self, name, val):
         try:
             assert (val in self._field_names) or (val is None)
         except AssertionError:
-            raise Exception(f"Invalid field name: {val}")
+            raise ValueError(f"Invalid field name: {val}")
 
     def _validate_all_field_names(self, name, val):
         try:
             for x in val:
                 self._validate_field_name(name, x)
         except AssertionError:
-            raise Exception("Fields must be a sequence of field names")
+            raise ValueError("Fields must be a sequence of field names")
 
     def _validate_single_char(self, name, val):
         try:
             assert _str_block_width(val) == 1
         except AssertionError:
-            raise Exception(f"Invalid value for {name}. Must be a string of length 1.")
+            raise ValueError(f"Invalid value for {name}. Must be a string of length 1.")
 
     def _validate_attributes(self, name, val):
         try:
             assert isinstance(val, dict)
         except AssertionError:
-            raise Exception("Attributes must be a dictionary of name/value pairs")
+            raise TypeError("Attributes must be a dictionary of name/value pairs")
 
     ##############################
     # ATTRIBUTE MANAGEMENT       #
     ##############################
     @property
-    def rows(self):
+    def rows(self) -> list[Any]:
         return self._rows[:]
 
     @property
-    def xhtml(self):
+    def xhtml(self) -> bool:
         """Print <br/> tags if True, <br> tags if False"""
         return self._xhtml
 
@@ -547,6 +574,22 @@ class PrettyTable:
         self._validate_option("xhtml", val)
         self._xhtml = val
 
+    @property
+    def none_format(self):
+        return self._none_format
+
+    @none_format.setter
+    def none_format(self, val):
+        if not self._field_names:
+            self._none_format = {}
+        elif val is None or (isinstance(val, dict) and len(val) == 0):
+            for field in self._field_names:
+                self._none_format[field] = None
+        else:
+            self._validate_none_format(val)
+            for field in self._field_names:
+                self._none_format[field] = val
+
     @property
     def field_names(self):
         """List or tuple of field names
@@ -570,6 +613,9 @@ class PrettyTable:
             for old_name in old_names:
                 if old_name not in self._align:
                     self._align.pop(old_name)
+        elif self._align:
+            for field_name in self._field_names:
+                self._align[field_name] = self._align[BASE_ALIGN_VALUE]
         else:
             self.align = "c"
         if self._valign and old_names:
@@ -591,15 +637,19 @@ class PrettyTable:
 
     @align.setter
     def align(self, val):
-        if not self._field_names:
-            self._align = {}
-        elif val is None or (isinstance(val, dict) and len(val) == 0):
-            for field in self._field_names:
-                self._align[field] = "c"
+        if val is None or (isinstance(val, dict) and len(val) == 0):
+            if not self._field_names:
+                self._align = {BASE_ALIGN_VALUE: "c"}
+            else:
+                for field in self._field_names:
+                    self._align[field] = "c"
         else:
             self._validate_align(val)
-            for field in self._field_names:
-                self._align[field] = val
+            if not self._field_names:
+                self._align = {BASE_ALIGN_VALUE: val}
+            else:
+                for field in self._field_names:
+                    self._align[field] = val
 
     @property
     def valign(self):
@@ -811,6 +861,21 @@ class PrettyTable:
         self._validate_option("border", val)
         self._border = val
 
+    @property
+    def preserve_internal_border(self):
+        """Controls printing of border inside table
+
+        Arguments:
+
+        preserve_internal_border - print a border inside the table even if
+            border is disabled (True or False)"""
+        return self._preserve_internal_border
+
+    @preserve_internal_border.setter
+    def preserve_internal_border(self, val):
+        self._validate_option("preserve_internal_border", val)
+        self._preserve_internal_border = val
+
     @property
     def hrules(self):
         """Controls printing of horizontal rules after rows
@@ -894,7 +959,7 @@ class PrettyTable:
             for field in self._field_names:
                 self._custom_format[field] = val
         else:
-            raise Exception(
+            raise TypeError(
                 "The custom_format property need to be a dictionary or callable"
             )
 
@@ -1064,7 +1129,8 @@ class PrettyTable:
 
     @property
     def top_right_junction_char(self):
-        """The character used when printing table borders to draw top-right line junctions
+        """
+        The character used when printing table borders to draw top-right line junctions
 
         Arguments:
 
@@ -1202,7 +1268,7 @@ class PrettyTable:
     # PRESET STYLE LOGIC         #
     ##############################
 
-    def set_style(self, style):
+    def set_style(self, style) -> None:
 
         if style == DEFAULT:
             self._set_default_style()
@@ -1221,7 +1287,7 @@ class PrettyTable:
         elif style == RANDOM:
             self._set_random_style()
         else:
-            raise Exception("Invalid pre-set style")
+            raise ValueError("Invalid pre-set style")
 
     def _set_orgmode_style(self):
         self._set_default_style()
@@ -1316,12 +1382,13 @@ class PrettyTable:
         self.vertical_char = random.choice(r"~!@#$%^&*()_+|-=\{}[];':\",./;<>?")
         self.horizontal_char = random.choice(r"~!@#$%^&*()_+|-=\{}[];':\",./;<>?")
         self.junction_char = random.choice(r"~!@#$%^&*()_+|-=\{}[];':\",./;<>?")
+        self.preserve_internal_border = random.choice((True, False))
 
     ##############################
     # DATA INPUT METHODS         #
     ##############################
 
-    def add_rows(self, rows):
+    def add_rows(self, rows) -> None:
 
         """Add rows to the table
 
@@ -1332,7 +1399,7 @@ class PrettyTable:
         for row in rows:
             self.add_row(row)
 
-    def add_row(self, row):
+    def add_row(self, row) -> None:
 
         """Add a row to the table
 
@@ -1342,7 +1409,7 @@ class PrettyTable:
         has fields"""
 
         if self._field_names and len(row) != len(self._field_names):
-            raise Exception(
+            raise ValueError(
                 "Row has incorrect number of values, "
                 f"(actual) {len(row)}!={len(self._field_names)} (expected)"
             )
@@ -1350,7 +1417,7 @@ class PrettyTable:
             self.field_names = [f"Field {n + 1}" for n in range(0, len(row))]
         self._rows.append(list(row))
 
-    def del_row(self, row_index):
+    def del_row(self, row_index) -> None:
 
         """Delete a row from the table
 
@@ -1359,13 +1426,15 @@ class PrettyTable:
         row_index - The index of the row you want to delete.  Indexing starts at 0."""
 
         if row_index > len(self._rows) - 1:
-            raise Exception(
+            raise IndexError(
                 f"Can't delete row at index {row_index}, "
                 f"table only has {len(self._rows)} rows"
             )
         del self._rows[row_index]
 
-    def add_column(self, fieldname, column, align="c", valign="t"):
+    def add_column(
+        self, fieldname, column, align: str = "c", valign: str = "t"
+    ) -> None:
 
         """Add a column to the table.
 
@@ -1390,12 +1459,12 @@ class PrettyTable:
                     self._rows.append([])
                 self._rows[i].append(column[i])
         else:
-            raise Exception(
+            raise ValueError(
                 f"Column length {len(column)} does not match number of rows "
                 f"{len(self._rows)}"
             )
 
-    def add_autoindex(self, fieldname="Index"):
+    def add_autoindex(self, fieldname: str = "Index"):
         """Add an auto-incrementing index column to the table.
         Arguments:
         fieldname - name of the field to contain the new column of data"""
@@ -1405,7 +1474,7 @@ class PrettyTable:
         for i, row in enumerate(self._rows):
             row.insert(0, i + 1)
 
-    def del_column(self, fieldname):
+    def del_column(self, fieldname) -> None:
 
         """Delete a column from the table
 
@@ -1414,7 +1483,7 @@ class PrettyTable:
         fieldname - The field name of the column you want to delete."""
 
         if fieldname not in self._field_names:
-            raise Exception(
+            raise ValueError(
                 "Can't delete column %r which is not a field name of this table."
                 " Field names are: %s"
                 % (fieldname, ", ".join(map(repr, self._field_names)))
@@ -1425,13 +1494,13 @@ class PrettyTable:
         for row in self._rows:
             del row[col_index]
 
-    def clear_rows(self):
+    def clear_rows(self) -> None:
 
         """Delete all rows from the table but keep the current field names"""
 
         self._rows = []
 
-    def clear(self):
+    def clear(self) -> None:
 
         """Delete all rows and field names from the table, maintaining nothing but
         styling options"""
@@ -1479,6 +1548,9 @@ class PrettyTable:
         for row in rows:
             for index, value in enumerate(row):
                 fieldname = self.field_names[index]
+                if self.none_format.get(fieldname) is not None:
+                    if value == "None" or value is None:
+                        value = self.none_format.get(fieldname)
                 if fieldname in self.max_width:
                     widths[index] = max(
                         widths[index],
@@ -1511,11 +1583,26 @@ class PrettyTable:
                 title_width = 0
             min_table_width = self.min_table_width or 0
             min_width = max(title_width, min_table_width)
-            table_width = self._compute_table_width(options)
-            if table_width < min_width:
+            if options["border"]:
+                borders = len(widths) + 1
+            elif options["preserve_internal_border"]:
+                borders = len(widths)
+            else:
+                borders = 0
+
+            # Subtract padding for each column and borders
+            min_width -= (
+                sum([sum(self._get_padding_widths(options)) for _ in widths]) + borders
+            )
+            # What is being scaled is content so we sum column widths
+            content_width = sum(widths) or 1
+
+            if content_width < min_width:
                 # Grow widths in proportion
-                scale = 1.0 * min_width / table_width
-                widths = [int(math.ceil(w * scale)) for w in widths]
+                scale = 1.0 * min_width / content_width
+                widths = [int(math.floor(w * scale)) for w in widths]
+                if sum(widths) < min_width:
+                    widths[-1] += min_width - sum(widths)
                 self._widths = widths
 
     def _get_padding_widths(self, options):
@@ -1572,7 +1659,7 @@ class PrettyTable:
     # PLAIN TEXT STRING METHODS  #
     ##############################
 
-    def get_string(self, **kwargs):
+    def get_string(self, **kwargs) -> str:
 
         """Return string representation of table in current state.
 
@@ -1584,6 +1671,8 @@ class PrettyTable:
         fields - names of fields (columns) to include
         header - print a header showing field names (True or False)
         border - print a border around the table (True or False)
+        preserve_internal_border - print a border inside the table even if
+            border is disabled (True or False)
         hrules - controls printing of horizontal rules after rows.
             Allowed values: ALL, FRAME, HEADER, NONE
         vrules - controls printing of vertical rules between columns.
@@ -1679,7 +1768,7 @@ class PrettyTable:
 
     def _stringify_hrule(self, options, where=""):
 
-        if not options["border"]:
+        if not options["border"] and not options["preserve_internal_border"]:
             return ""
         lpad, rpad = self._get_padding_widths(options)
         if options["vrules"] in (ALL, FRAME):
@@ -1711,6 +1800,10 @@ class PrettyTable:
         if options["vrules"] in (ALL, FRAME):
             bits.pop()
             bits.append(options[where + "right_junction_char"])
+
+        if options["preserve_internal_border"] and not options["border"]:
+            bits = bits[1:-1]
+
         return "".join(bits)
 
     def _stringify_title(self, title, options):
@@ -1726,7 +1819,9 @@ class PrettyTable:
                 lines.append(self._stringify_hrule(options, "top_"))
         bits = []
         endpoint = (
-            options["vertical_char"] if options["vrules"] in (ALL, FRAME) else " "
+            options["vertical_char"]
+            if options["vrules"] in (ALL, FRAME) and options["border"]
+            else " "
         )
         bits.append(endpoint)
         title = " " * lpad + title + " " * rpad
@@ -1772,22 +1867,32 @@ class PrettyTable:
                 fieldname = field.lower()
             else:
                 fieldname = field
+            if _str_block_width(fieldname) > width:
+                fieldname = fieldname[:width]
             bits.append(
                 " " * lpad
                 + self._justify(fieldname, width, self._align[field])
                 + " " * rpad
             )
-            if options["border"]:
+            if options["border"] or options["preserve_internal_border"]:
                 if options["vrules"] == ALL:
                     bits.append(options["vertical_char"])
                 else:
                     bits.append(" ")
+
+        # If only preserve_internal_border is true, then we just appended
+        # a vertical character at the end when we wanted a space
+        if not options["border"] and options["preserve_internal_border"]:
+            bits.pop()
+            bits.append(" ")
         # If vrules is FRAME, then we just appended a space at the end
         # of the last field, when we really want a vertical character
         if options["border"] and options["vrules"] == FRAME:
             bits.pop()
             bits.append(options["vertical_char"])
-        if options["border"] and options["hrules"] != NONE:
+        if (options["border"] or options["preserve_internal_border"]) and options[
+            "hrules"
+        ] != NONE:
             bits.append("\n")
             bits.append(self._hrule)
         return "".join(bits)
@@ -1801,6 +1906,8 @@ class PrettyTable:
             lines = value.split("\n")
             new_lines = []
             for line in lines:
+                if line == "None" and self.none_format.get(field) is not None:
+                    line = self.none_format[field]
                 if _str_block_width(line) > width:
                     line = textwrap.fill(line, width)
                 new_lines.append(line)
@@ -1851,13 +1958,19 @@ class PrettyTable:
                     + self._justify(line, width, self._align[field])
                     + " " * rpad
                 )
-                if options["border"]:
+                if options["border"] or options["preserve_internal_border"]:
                     if options["vrules"] == ALL:
                         bits[y].append(self.vertical_char)
                     else:
                         bits[y].append(" ")
                 y += 1
 
+        # If only preserve_internal_border is true, then we just appended
+        # a vertical character at the end when we wanted a space
+        if not options["border"] and options["preserve_internal_border"]:
+            bits[-1].pop()
+            bits[-1].append(" ")
+
         # If vrules is FRAME, then we just appended a space at the end
         # of the last field, when we really want a vertical character
         for y in range(0, row_height):
@@ -1874,7 +1987,7 @@ class PrettyTable:
 
         return "\n".join(bits)
 
-    def paginate(self, page_length=58, **kwargs):
+    def paginate(self, page_length: int = 58, line_break: str = "\f", **kwargs):
 
         pages = []
         kwargs["start"] = kwargs.get("start", 0)
@@ -1885,12 +1998,12 @@ class PrettyTable:
             if kwargs["end"] == true_end:
                 break
             kwargs["start"] += page_length
-        return "\f".join(pages)
+        return line_break.join(pages)
 
     ##############################
     # CSV STRING METHODS         #
     ##############################
-    def get_csv_string(self, **kwargs):
+    def get_csv_string(self, **kwargs) -> str:
 
         """Return string representation of CSV formatted table in the current state
 
@@ -1918,7 +2031,7 @@ class PrettyTable:
     ##############################
     # JSON STRING METHODS        #
     ##############################
-    def get_json_string(self, **kwargs):
+    def get_json_string(self, **kwargs) -> str:
 
         """Return string representation of JSON formatted table in the current state
 
@@ -1930,7 +2043,7 @@ class PrettyTable:
         """
 
         options = self._get_options(kwargs)
-        json_options = dict(indent=4, separators=(",", ": "), sort_keys=True)
+        json_options: Any = dict(indent=4, separators=(",", ": "), sort_keys=True)
         json_options.update(
             {key: value for key, value in kwargs.items() if key not in options}
         )
@@ -1947,7 +2060,7 @@ class PrettyTable:
     # HTML STRING METHODS        #
     ##############################
 
-    def get_html_string(self, **kwargs):
+    def get_html_string(self, **kwargs) -> str:
         """Return string representation of HTML formatted version of table in current
         state.
 
@@ -1959,6 +2072,8 @@ class PrettyTable:
         fields - names of fields (columns) to include
         header - print a header showing field names (True or False)
         border - print a border around the table (True or False)
+        preserve_internal_border - print a border inside the table even if
+            border is disabled (True or False)
         hrules - controls printing of horizontal rules after rows.
             Allowed values: ALL, FRAME, HEADER, NONE
         vrules - controls printing of vertical rules between columns.
@@ -2005,12 +2120,7 @@ class PrettyTable:
         # Title
         title = options["title"] or self._title
         if title:
-            cols = (
-                len(options["fields"]) if options["fields"] else len(self.field_names)
-            )
-            lines.append("    <tr>")
-            lines.append(f"        <td colspan={cols}>{title}</td>")
-            lines.append("    </tr>")
+            lines.append(f"    <caption>{title}</caption>")
 
         # Headers
         if options["header"]:
@@ -2068,6 +2178,8 @@ class PrettyTable:
                 open_tag.append(' frame="vsides"')
             elif options["vrules"] == ALL:
                 open_tag.append(' frame="vsides" rules="cols"')
+        if not options["border"] and options["preserve_internal_border"]:
+            open_tag.append(' rules="cols"')
         if options["attributes"]:
             for attr_name in options["attributes"]:
                 open_tag.append(f' {attr_name}="{options["attributes"][attr_name]}"')
@@ -2077,12 +2189,7 @@ class PrettyTable:
         # Title
         title = options["title"] or self._title
         if title:
-            cols = (
-                len(options["fields"]) if options["fields"] else len(self.field_names)
-            )
-            lines.append("    <tr>")
-            lines.append(f"        <td colspan={cols}>{title}</td>")
-            lines.append("    </tr>")
+            lines.append(f"    <caption>{title}</caption>")
 
         # Headers
         if options["header"]:
@@ -2138,7 +2245,7 @@ class PrettyTable:
     # LATEX STRING METHODS       #
     ##############################
 
-    def get_latex_string(self, **kwargs):
+    def get_latex_string(self, **kwargs) -> str:
         """Return string representation of LaTex formatted version of table in current
         state.
 
@@ -2149,6 +2256,8 @@ class PrettyTable:
         fields - names of fields (columns) to include
         header - print a header showing field names (True or False)
         border - print a border around the table (True or False)
+        preserve_internal_border - print a border inside the table even if
+            border is disabled (True or False)
         hrules - controls printing of horizontal rules after rows.
             Allowed values: ALL, FRAME, HEADER, NONE
         vrules - controls printing of vertical rules between columns.
@@ -2215,6 +2324,8 @@ class PrettyTable:
         wanted_alignments = [self._align[field] for field in wanted_fields]
         if options["border"] and options["vrules"] == ALL:
             alignment_str = "|".join(wanted_alignments)
+        elif not options["border"] and options["preserve_internal_border"]:
+            alignment_str = "|".join(wanted_alignments)
         else:
             alignment_str = "".join(wanted_alignments)
 
@@ -2223,14 +2334,15 @@ class PrettyTable:
 
         begin_cmd = "\\begin{tabular}{%s}" % alignment_str
         lines.append(begin_cmd)
-
         if options["border"] and options["hrules"] in [ALL, FRAME]:
             lines.append("\\hline")
 
         # Headers
         if options["header"]:
             lines.append(" & ".join(wanted_fields) + " \\\\")
-        if options["border"] and options["hrules"] in [ALL, HEADER]:
+        if (options["border"] or options["preserve_internal_border"]) and options[
+            "hrules"
+        ] in [ALL, HEADER]:
             lines.append("\\hline")
 
         # Data
@@ -2267,7 +2379,7 @@ def _str_block_width(val):
 ##############################
 
 
-def from_csv(fp, field_names=None, **kwargs):
+def from_csv(fp, field_names: Any | None = None, **kwargs):
     fmtparams = {}
     for param in [
         "delimiter",
@@ -2320,19 +2432,19 @@ def from_json(json_string, **kwargs):
 
 
 class TableHandler(HTMLParser):
-    def __init__(self, **kwargs):
+    def __init__(self, **kwargs) -> None:
         HTMLParser.__init__(self)
         self.kwargs = kwargs
-        self.tables = []
-        self.last_row = []
-        self.rows = []
+        self.tables: list[list] = []
+        self.last_row: list[str] = []
+        self.rows: list[Any] = []
         self.max_row_width = 0
         self.active = None
         self.last_content = ""
         self.is_last_row_header = False
         self.colspan = 0
 
-    def handle_starttag(self, tag, attrs):
+    def handle_starttag(self, tag, attrs) -> None:
         self.active = tag
         if tag == "th":
             self.is_last_row_header = True
@@ -2340,7 +2452,7 @@ class TableHandler(HTMLParser):
             if key == "colspan":
                 self.colspan = int(value)
 
-    def handle_endtag(self, tag):
+    def handle_endtag(self, tag) -> None:
         if tag in ["th", "td"]:
             stripped_content = self.last_content.strip()
             self.last_row.append(stripped_content)
@@ -2361,7 +2473,7 @@ class TableHandler(HTMLParser):
         self.last_content = " "
         self.active = None
 
-    def handle_data(self, data):
+    def handle_data(self, data) -> None:
         self.last_content += data
 
     def generate_table(self, rows):
@@ -2382,7 +2494,7 @@ class TableHandler(HTMLParser):
                 table.add_row(row[0])
         return table
 
-    def make_fields_unique(self, fields):
+    def make_fields_unique(self, fields) -> None:
         """
         iterates over the row and make each field unique
         """
@@ -2413,7 +2525,7 @@ def from_html_one(html_code, **kwargs):
     try:
         assert len(tables) == 1
     except AssertionError:
-        raise Exception(
+        raise ValueError(
             "More than one <table> in provided HTML code. Use from_html instead."
         )
     return tables[0]
