@@ -1,9 +1,11 @@
+from __future__ import annotations
+
 import datetime as dt
 import io
 import random
 import sqlite3
 from math import e, pi, sqrt
-from typing import Any, List
+from typing import Any
 
 import pytest
 
@@ -94,6 +96,119 @@ def mix_prettytable():
     return mix
 
 
+class TestNoneOption:
+    def test_none_char_valid_option(self):
+        PrettyTable(["Field 1", "Field 2", "Field 3"], none_format="")
+
+    def test_none_char_invalid_option(self):
+        with pytest.raises(TypeError) as exc:
+            PrettyTable(["Field 1", "Field 2", "Field 3"], none_format=2)
+            assert "must be a string." in exc.value
+
+    def test_no_value_replace_none(self):
+        t = PrettyTable(["Field 1", "Field 2", "Field 3"])
+        t.add_row(["value 1", None, "value 2"])
+        assert (
+            t.get_string().strip()
+            == """
++---------+---------+---------+
+| Field 1 | Field 2 | Field 3 |
++---------+---------+---------+
+| value 1 |   None  | value 2 |
++---------+---------+---------+
+""".strip()
+        )
+
+    def test_no_value_replace_none_with_default_field_names(self):
+        t = PrettyTable()
+        t.add_row(["value 1", "None", "value 2"])
+        assert (
+            t.get_string().strip()
+            == """
++---------+---------+---------+
+| Field 1 | Field 2 | Field 3 |
++---------+---------+---------+
+| value 1 |   None  | value 2 |
++---------+---------+---------+
+""".strip()
+        )
+
+    def test_replace_none_all(self):
+        t = PrettyTable(["Field 1", "Field 2", "Field 3"], none_format="N/A")
+        t.add_row(["value 1", None, "None"])
+        assert (
+            t.get_string().strip()
+            == """
++---------+---------+---------+
+| Field 1 | Field 2 | Field 3 |
++---------+---------+---------+
+| value 1 |   N/A   |   N/A   |
++---------+---------+---------+
+""".strip()
+        )
+
+    def test_replace_none_by_col(self):
+        t = PrettyTable(["Field 1", "Field 2", "Field 3"])
+        t.none_format["Field 2"] = "N/A"
+        t.none_format["Field 3"] = ""
+        t.add_row(["value 1", None, None])
+        assert (
+            t.get_string().strip()
+            == """
++---------+---------+---------+
+| Field 1 | Field 2 | Field 3 |
++---------+---------+---------+
+| value 1 |   N/A   |         |
++---------+---------+---------+
+""".strip()
+        )
+
+    def test_replace_none_recompute_width(self):
+        t = PrettyTable()
+        t.add_row([None])
+        t.none_format = "0123456789"
+        assert (
+            t.get_string().strip()
+            == """
++------------+
+|  Field 1   |
++------------+
+| 0123456789 |
++------------+
+""".strip()
+        )
+
+    def test_replace_none_maintain_width_on_recompute(self):
+        t = PrettyTable()
+        t.add_row(["Hello"])
+        t.none_format = "0123456789"
+        assert (
+            t.get_string().strip()
+            == """
++---------+
+| Field 1 |
++---------+
+|  Hello  |
++---------+
+""".strip()
+        )
+
+    def test_replace_none_recompute_width_multi_column(self):
+        t = PrettyTable()
+        t.add_row(["Hello", None, "World"])
+        t.none_format = "0123456789"
+        assert (
+            t.get_string().strip()
+            == """
++---------+------------+---------+
+| Field 1 |  Field 2   | Field 3 |
++---------+------------+---------+
+|  Hello  | 0123456789 |  World  |
++---------+------------+---------+
+""".strip()
+        )
+
+
 class TestBuildEquivalence:
     """Make sure that building a table row-by-row and column-by-column yield the same
     results"""
@@ -161,11 +276,11 @@ class TestDeleteColumn:
 
         assert with_del.get_string() == without_row.get_string()
 
-    def test_delete_illegal_column_raises_exception(self):
+    def test_delete_illegal_column_raises_error(self):
         table = PrettyTable()
         table.add_column("City name", ["Adelaide", "Brisbane", "Darwin"])
 
-        with pytest.raises(Exception):
+        with pytest.raises(ValueError):
             table.del_column("City not-a-name")
 
 
@@ -214,6 +329,62 @@ class TestFieldNameLessTable:
         )
 
 
+@pytest.fixture(scope="function")
+def aligned_before_table():
+    x = PrettyTable()
+    x.align = "r"
+    x.field_names = ["City name", "Area", "Population", "Annual Rainfall"]
+    x.add_row(["Adelaide", 1295, 1158259, 600.5])
+    x.add_row(["Brisbane", 5905, 1857594, 1146.4])
+    x.add_row(["Darwin", 112, 120900, 1714.7])
+    x.add_row(["Hobart", 1357, 205556, 619.5])
+    x.add_row(["Sydney", 2058, 4336374, 1214.8])
+    x.add_row(["Melbourne", 1566, 3806092, 646.9])
+    x.add_row(["Perth", 5386, 1554769, 869.4])
+    return x
+
+
+@pytest.fixture(scope="function")
+def aligned_after_table():
+    x = PrettyTable()
+    x.field_names = ["City name", "Area", "Population", "Annual Rainfall"]
+    x.add_row(["Adelaide", 1295, 1158259, 600.5])
+    x.add_row(["Brisbane", 5905, 1857594, 1146.4])
+    x.add_row(["Darwin", 112, 120900, 1714.7])
+    x.add_row(["Hobart", 1357, 205556, 619.5])
+    x.add_row(["Sydney", 2058, 4336374, 1214.8])
+    x.add_row(["Melbourne", 1566, 3806092, 646.9])
+    x.add_row(["Perth", 5386, 1554769, 869.4])
+    x.align = "r"
+    return x
+
+
+class TestAlignment:
+
+    """Make sure alignment works regardless of when it was set"""
+
+    def test_aligned_ascii(
+        self, aligned_before_table: prettytable, aligned_after_table: prettytable
+    ):
+        before = aligned_before_table.get_string()
+        after = aligned_after_table.get_string()
+        assert before == after
+
+    def test_aligned_html(
+        self, aligned_before_table: prettytable, aligned_after_table: prettytable
+    ):
+        before = aligned_before_table.get_html_string()
+        after = aligned_after_table.get_html_string()
+        assert before == after
+
+    def test_aligned_latex(
+        self, aligned_before_table: prettytable, aligned_after_table: prettytable
+    ):
+        before = aligned_before_table.get_latex_string()
+        after = aligned_after_table.get_latex_string()
+        assert before == after
+
+
 @pytest.fixture(scope="function")
 def city_data_prettytable():
     """Just build the Australian capital city data example table."""
@@ -315,6 +486,10 @@ class TestOptionAttribute:
             city_data_prettytable.get_string() == city_data_prettytable[:].get_string()
         )
 
+    def test_preserve_internal_border(self):
+        pt = PrettyTable(preserve_internal_border=True)
+        assert pt.preserve_internal_border is True
+
 
 @pytest.fixture(scope="module")
 def db_cursor():
@@ -380,6 +555,13 @@ class TestBasic:
         city_data_prettytable.title = "My table"
         self._test_all_length_equal(city_data_prettytable)
 
+    def test_all_lengths_equal_with_long_title(
+        self, city_data_prettytable: PrettyTable
+    ):
+        """All lines in a table should be of the same length, even with a long title."""
+        city_data_prettytable.title = "My table (75 characters wide) " + "=" * 45
+        self._test_all_length_equal(city_data_prettytable)
+
     def test_no_blank_lines_without_border(self, city_data_prettytable: PrettyTable):
         """No table should ever have blank lines in it."""
         city_data_prettytable.border = False
@@ -688,7 +870,7 @@ class TestBreakLine:
         ],
     )
     def test_break_line_ASCII(
-        self, rows: List[List[Any]], hrule: int, expected_result: str
+        self, rows: list[list[Any]], hrule: int, expected_result: str
     ):
         t = PrettyTable(["Field 1", "Field 2"])
         for row in rows:
@@ -883,6 +1065,80 @@ class TestHtmlOutput:
         </tr>
     </tbody>
 </table>
+""".strip()  # noqa: E501
+        )
+
+    def test_HtmlOutputWithTitle(self):
+        t = helper_table()
+        t.title = "Title"
+        result = t.get_html_string()
+        assert (
+            result.strip()
+            == """
+<table>
+    <caption>Title</caption>
+    <thead>
+        <tr>
+            <th>Field 1</th>
+            <th>Field 2</th>
+            <th>Field 3</th>
+        </tr>
+    </thead>
+    <tbody>
+        <tr>
+            <td>value 1</td>
+            <td>value2</td>
+            <td>value3</td>
+        </tr>
+        <tr>
+            <td>value 4</td>
+            <td>value5</td>
+            <td>value6</td>
+        </tr>
+        <tr>
+            <td>value 7</td>
+            <td>value8</td>
+            <td>value9</td>
+        </tr>
+    </tbody>
+</table>
+""".strip()
+        )
+
+    def test_HtmlOutputFormattedWithTitle(self):
+        t = helper_table()
+        t.title = "Title"
+        result = t.get_html_string(format=True)
+        assert (
+            result.strip()
+            == """
+<table frame="box" rules="cols">
+    <caption>Title</caption>
+    <thead>
+        <tr>
+            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Field 1</th>
+            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Field 2</th>
+            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Field 3</th>
+        </tr>
+    </thead>
+    <tbody>
+        <tr>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value 1</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value2</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value3</td>
+        </tr>
+        <tr>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value 4</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value5</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value6</td>
+        </tr>
+        <tr>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value 7</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value8</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value9</td>
+        </tr>
+    </tbody>
+</table>
 """.strip()  # noqa: E501
         )
 
@@ -1091,12 +1347,10 @@ class TestStyle:
             ),
             pytest.param(
                 PLAIN_COLUMNS,
-                """
-Field 1        Field 2        Field 3        
-value 1         value2         value3        
-value 4         value5         value6        
-value 7         value8         value9
-""",  # noqa: W291
+                "Field 1        Field 2        Field 3        \n"
+                "value 1         value2         value3        \n"
+                "value 4         value5         value6        \n"
+                "value 7         value8         value9",
                 id="PLAIN_COLUMNS",
             ),
             pytest.param(
@@ -1156,7 +1410,7 @@ value 7         value8         value9
 
         # Act / Assert
         # This is an hrule style, not a table style
-        with pytest.raises(Exception):
+        with pytest.raises(ValueError):
             t.set_style(ALL)
 
     @pytest.mark.parametrize(
@@ -1314,7 +1568,7 @@ class TestHtmlConstructor:
     def test_HtmlOneFailOnMany(self, city_data_prettytable: PrettyTable):
         html_string = city_data_prettytable.get_html_string()
         html_string += city_data_prettytable.get_html_string()
-        with pytest.raises(Exception):
+        with pytest.raises(ValueError):
             from_html_one(html_string)
 
 
@@ -1454,6 +1708,13 @@ def test_paginate():
     assert "\f" in paginated
     assert paginated.endswith(expected_page_2)
 
+    # Act
+    paginated = t.paginate(page_length=4, line_break="\n")
+
+    # Assert
+    assert "\f" not in paginated
+    assert "\n" in paginated
+
 
 def test_add_rows():
     """A table created with multiple add_row calls
@@ -1551,7 +1812,7 @@ class TestCustomFormatter:
         assert len(pt.custom_format) == 1
 
     def test_init_custom_format_throw_error_is_not_callable(self):
-        with pytest.raises(Exception) as e:
+        with pytest.raises(ValueError) as e:
             PrettyTable(custom_format={"col1": "{:.2}"})
 
         assert "Invalid value for custom_format.col1. Must be a function." in str(
@@ -1571,7 +1832,7 @@ class TestCustomFormatter:
 
     def test_set_custom_format_invalid_type_throw_error(self):
         pt = PrettyTable()
-        with pytest.raises(Exception) as e:
+        with pytest.raises(TypeError) as e:
             pt.custom_format = "Some String"
         assert "The custom_format property need to be a dictionary or callable" in str(
             e.value
@@ -1649,3 +1910,145 @@ class TestRepr:
 
     def test_jupyter_repr(self, row_prettytable: PrettyTable):
         assert row_prettytable._repr_html_() == row_prettytable.get_html_string()
+
+
+class TestMinTableWidth:
+    @pytest.mark.parametrize(
+        "loops, fields, desired_width, border, internal_border",
+        [
+            (15, ["Test table"], 20, True, False),
+            (16, ["Test table"], 21, True, False),
+            (18, ["Test table", "Test table 2"], 40, True, False),
+            (19, ["Test table", "Test table 2"], 41, True, False),
+            (21, ["Test table", "Test col 2", "Test col 3"], 50, True, False),
+            (22, ["Test table", "Test col 2", "Test col 3"], 51, True, False),
+            (19, ["Test table"], 20, False, False),
+            (20, ["Test table"], 21, False, False),
+            (25, ["Test table", "Test table 2"], 40, False, False),
+            (26, ["Test table", "Test table 2"], 41, False, False),
+            (25, ["Test table", "Test col 2", "Test col 3"], 50, False, False),
+            (26, ["Test table", "Test col 2", "Test col 3"], 51, False, False),
+            (18, ["Test table"], 20, False, True),
+            (19, ["Test table"], 21, False, True),
+            (23, ["Test table", "Test table 2"], 40, False, True),
+            (24, ["Test table", "Test table 2"], 41, False, True),
+            (22, ["Test table", "Test col 2", "Test col 3"], 50, False, True),
+            (23, ["Test table", "Test col 2", "Test col 3"], 51, False, True),
+        ],
+    )
+    def test_min_table_width(
+        self, loops, fields, desired_width, border, internal_border
+    ):
+        for col_width in range(loops):
+            x = prettytable.PrettyTable()
+            x.border = border
+            x.preserve_internal_border = internal_border
+            x.field_names = fields
+            x.add_row(["X" * col_width] + ["" for _ in range(len(fields) - 1)])
+            x.min_table_width = desired_width
+            t = x.get_string()
+            if border is False and internal_border is False:
+                assert [len(x) for x in t.split("\n")] == [desired_width, desired_width]
+            elif border is False and internal_border is True:
+                assert [len(x) for x in t.split("\n")] == [
+                    desired_width,
+                    desired_width - 1,
+                    desired_width,
+                ]
+            else:
+                assert [len(x) for x in t.split("\n")] == [
+                    desired_width,
+                    desired_width,
+                    desired_width,
+                    desired_width,
+                    desired_width,
+                ]
+
+
+class TestMaxTableWidth:
+    def test_max_table_width(self):
+        pt = PrettyTable()
+        pt.max_table_width = 5
+        pt.add_row([0])
+
+        assert (
+            pt.get_string().strip()
+            == """
++-----+
+| Fie |
++-----+
+|  0  |
++-----+
+""".strip()
+        )
+
+
+class TestPreservingInternalBorders:
+    def test_internal_border_preserved(self):
+        pt = helper_table(3)
+        pt.border = False
+        pt.preserve_internal_border = True
+
+        assert (
+            pt.get_string().strip()
+            == """
+ Field 1 | Field 2 | Field 3  
+---------+---------+---------
+ value 1 |  value2 |  value3  
+ value 4 |  value5 |  value6  
+ value 7 |  value8 |  value9 
+""".strip()  # noqa: W291
+        )
+
+    def test_internal_border_preserved_latex(self):
+        pt = helper_table(3)
+        pt.border = False
+        pt.format = True
+        pt.preserve_internal_border = True
+
+        assert pt.get_latex_string().strip() == (
+            "\\begin{tabular}{c|c|c}\r\n"
+            "Field 1 & Field 2 & Field 3 \\\\\r\n"
+            "value 1 & value2 & value3 \\\\\r\n"
+            "value 4 & value5 & value6 \\\\\r\n"
+            "value 7 & value8 & value9 \\\\\r\n"
+            "\\end{tabular}"
+        )
+
+    def test_internal_border_preserved_html(self):
+        pt = helper_table(3)
+        pt.format = True
+        pt.border = False
+        pt.preserve_internal_border = True
+
+        assert (
+            pt.get_html_string().strip()
+            == """
+<table rules="cols">
+    <thead>
+        <tr>
+            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Field 1</th>
+            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Field 2</th>
+            <th style="padding-left: 1em; padding-right: 1em; text-align: center">Field 3</th>
+        </tr>
+    </thead>
+    <tbody>
+        <tr>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value 1</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value2</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value3</td>
+        </tr>
+        <tr>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value 4</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value5</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value6</td>
+        </tr>
+        <tr>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value 7</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value8</td>
+            <td style="padding-left: 1em; padding-right: 1em; text-align: center; vertical-align: top">value9</td>
+        </tr>
+    </tbody>
+</table>
+""".strip()  # noqa: E501
+        )
